@@ -1,8 +1,7 @@
 /*
  * The parse tree transformation module for SIP.
  *
- * Copyright (c) 2008
- * 	Phil Thompson <phil@river-bank.demon.co.uk>
+ * Copyright (c) 2008 Riverbank Computing Limited <info@riverbankcomputing.com>
  * 
  * This file is part of SIP.
  * 
@@ -81,6 +80,7 @@ static void registerMetaType(classDef *cd);
 static void addComplementarySlots(sipSpec *pt, classDef *cd);
 static void addComplementarySlot(sipSpec *pt, classDef *cd, memberDef *md,
         slotType cslot, const char *cslot_name);
+static void resolveInstantiatedClassTemplate(sipSpec *pt, argDef *type);
 
 
 /*
@@ -400,8 +400,7 @@ static void registerMetaType(classDef *cd)
 
     /*
      * We register types with Qt if the class is not abstract, has a public
-     * default ctor, a public copy ctor, a public dtor and isn't one of the
-     * internally supported types.
+     * default ctor, a public copy ctor and a public dtor.
      */
     if (isAbstractClass(cd))
         return;
@@ -409,18 +408,6 @@ static void registerMetaType(classDef *cd)
     if (!isPublicDtor(cd))
         return;
 
-    if (classFQCName(cd)->next == NULL)
-    {
-        if (strcmp(classBaseName(cd), "QChar") == 0)
-            return;
-
-        if (strcmp(classBaseName(cd), "QString") == 0)
-            return;
-
-        if (strcmp(classBaseName(cd), "QByteArray") == 0)
-            return;
-    }
-
     pub_def_ctor = pub_copy_ctor = FALSE;
 
     for (ct = cd->ctors; ct != NULL; ct = ct->next)
@@ -830,7 +817,6 @@ static classDef *getProxy(moduleDef *mod, classDef *cd)
  */
 static void filterMainModuleVirtualHandlers(moduleDef *mod)
 {
-    int nr = 0;
     moduleListDef *mld;
     virtHandlerDef *vhd;
 
@@ -839,12 +825,8 @@ static void filterMainModuleVirtualHandlers(moduleDef *mod)
      * those at the deepest level of %Import, are done first.
      */
     for (mld = mod->allimports; mld != NULL; mld = mld->next)
-    {
-        mld->module->modulenr = nr++;
         filterModuleVirtualHandlers(mld->module);
-    }
 
-    mod->modulenr = nr;
     filterModuleVirtualHandlers(mod);
 
     /*
@@ -1027,8 +1009,13 @@ static void setHierarchy(sipSpec *pt,classDef *base,classDef *cd,
     if (cd -> mro != NULL)
         return;
 
-    if (cd -> ecd != NULL)
-        setHierarchy(pt,base,cd -> ecd,head);
+    if (cd->ecd != NULL)
+    {
+        setHierarchy(pt, base, cd->ecd, head);
+
+        if (isDeprecatedClass(cd->ecd))
+            setIsDeprecatedClass(cd);
+    }
 
     if (cd -> iff -> type == class_iface)
     {
@@ -1055,6 +1042,9 @@ static void setHierarchy(sipSpec *pt,classDef *base,classDef *cd,
             {
                 appendToMRO(cd -> mro,&tailp,mro -> cd);
 
+                if (isDeprecatedClass(mro->cd))
+                    setIsDeprecatedClass(cd);
+
                 /*
                  * If the super-class is a QObject sub-class then this one is
                  * as well.
@@ -1204,6 +1194,9 @@ static void transformCtors(sipSpec *pt, classDef *cd)
                 fatalScopedName(classFQCName(cd));
                 fatal(" has ctors with the same Python signature\n");
             }
+
+        if (isDeprecatedClass(cd))
+            setIsDeprecatedCtor(ct);
     }
 }
 
@@ -1295,12 +1288,19 @@ static void addDefaultCopyCtor(classDef *cd)
         copyct -> posthook = NULL;
         copyct -> next = NULL;
  
+        if (isDeprecatedClass(cd))
+            setIsDeprecatedCtor(copyct);
+
         /* Append it to the list. */
         for (tailp = &cd -> ctors; *tailp != NULL; tailp = &(*tailp) -> next)
             ;
  
         *tailp = copyct;
     }
+
+    /* We assume it has an assignment operator if it has a public copy ctor. */
+    if (isPublicCtor(copyct))
+        setCanAssign(cd);
 }
 
 
@@ -1338,6 +1338,9 @@ static void transformScopeOverloads(sipSpec *pt, classDef *scope,
                 fatal("%s() has overloaded functions with the same Python signature\n", od->common->pyname->text);
             }
         }
+
+        if (scope != NULL && isDeprecatedClass(scope))
+            setIsDeprecated(od);
     }
 }
 
@@ -1350,7 +1353,8 @@ static void transformVariableList(sipSpec *pt)
     varDef *vd;
 
     for (vd = pt->vars; vd != NULL; vd = vd->next)
-        resolveVariableType(pt, vd);
+        if (vd->ecd == NULL || !isTemplateClass(vd->ecd))
+            resolveVariableType(pt, vd);
 }
 
 
@@ -1830,12 +1834,8 @@ static void resolveVariableType(sipSpec *pt, varDef *vd)
     else
         ifaceFileIsUsed(&vd->module->used, vtype);
 
-    /*
-     * Instance variables or static class variables (unless they are constants)
-     * need a handler.
-     */
-    if (vd->ecd != NULL && vd->accessfunc == NULL &&
-        (!isStaticVar(vd) || vtype->nrderefs != 0 || !isConstArg(vtype)))
+    /* Scoped variables need a handler unless they have %AccessCode. */
+    if (vd->ecd != NULL && vd->accessfunc == NULL)
     {
         setNeedsHandler(vd);
         setHasVarHandlers(vd->ecd);
@@ -2176,18 +2176,20 @@ int sameSignature(signatureDef *sd1,signatureDef *sd2,int strict)
 
 
 #define pyAsString(t)   ((t) == ustring_type || (t) == sstring_type || \
-             (t) == string_type)
+            (t) == string_type)
 #define pyAsFloat(t)    ((t) == cfloat_type || (t) == float_type || \
-             (t) == cdouble_type || (t) == double_type)
-#define pyAsInt(t)  ((t) == cint_type || (t) == bool_type || \
-             (t) == short_type || (t) == ushort_type || \
-             (t) == int_type || (t) == uint_type)
+            (t) == cdouble_type || (t) == double_type)
+#define pyAsInt(t)  ((t) == bool_type || \
+            (t) == short_type || (t) == ushort_type || \
+            (t) == cint_type || (t) == int_type || (t) == uint_type)
 #define pyAsLong(t) ((t) == long_type || (t) == longlong_type)
 #define pyAsULong(t)    ((t) == ulong_type || (t) == ulonglong_type)
 #define pyAsAuto(t) ((t) == bool_type || \
-             (t) == short_type || (t) == ushort_type || \
-             (t) == int_type || (t) == uint_type || \
-             (t) == float_type || (t) == double_type)
+            (t) == short_type || (t) == ushort_type || \
+            (t) == int_type || (t) == uint_type || \
+            (t) == float_type || (t) == double_type)
+#define pyIsConstrained(t)  ((t) == cbool_type || (t) == cint_type || \
+            (t) == cfloat_type || (t) == cdouble_type)
 
 /*
  * Compare two argument types and return TRUE if they are the same.  "strict"
@@ -2208,6 +2210,15 @@ static int sameArgType(argDef *a1, argDef *a2, int strict)
         return sameBaseType(a1,a2);
     }
 
+    /* If both are constrained fundamental types then the types must match. */
+    if (pyIsConstrained(a1->atype) && pyIsConstrained(a2->atype))
+        return (a1->atype == a2->atype);
+
+    /* An unconstrained enum also acts as a (very) constrained int. */
+    if ((pyAsInt(a1->atype) && a2->atype == enum_type && !isConstrained(a2)) ||
+        (a1->atype == enum_type && !isConstrained(a1) && pyAsInt(a2->atype)))
+        return TRUE;
+
     /* Python will see all these as strings. */
     if (pyAsString(a1->atype) && pyAsString(a2->atype))
         return TRUE;
@@ -2240,13 +2251,24 @@ static int sameArgType(argDef *a1, argDef *a2, int strict)
 /*
  * Compare two basic types and return TRUE if they are the same.
  */
-
-int sameBaseType(argDef *a1,argDef *a2)
+int sameBaseType(argDef *a1, argDef *a2)
 {
     /* The types must be the same. */
+    if (a1->atype != a2->atype)
+    {
+        /*
+         * If we are comparing a template with those that have already been
+         * used to instantiate a class then we need to compare with the class
+         * name.  Hopefully this won't have wider side effects.
+         */
+        if (a1->atype == class_type && a2->atype == defined_type)
+            return sameScopedName(a1->u.cd->iff->fqcname, a2->u.snd);
+
+        if (a1->atype == defined_type && a2->atype == class_type)
+            return sameScopedName(a1->u.snd, a2->u.cd->iff->fqcname);
 
-    if (a1 -> atype != a2 ->atype)
         return FALSE;
+    }
 
     switch (a1 -> atype)
     {
@@ -2524,21 +2546,7 @@ static void getBaseType(sipSpec *pt, moduleDef *mod, classDef *defscope, argDef
     }
 
     /* See if the type refers to an instantiated template. */
-    if (type->atype == template_type)
-    {
-        classDef *cd;
-
-        for (cd = pt->classes; cd != NULL; cd = cd->next)
-            if (cd->td != NULL &&
-                sameScopedName(cd->td->fqname, type->u.td->fqname) &&
-                sameSignature(&cd->td->types, &type->u.td->types, TRUE))
-            {
-                type->atype = class_type;
-                type->u.cd = cd;
-
-                break;
-            }
-    }
+    resolveInstantiatedClassTemplate(pt, type);
 
     /* Replace the base type if it has been mapped. */
     if (type -> atype == struct_type || type -> atype == template_type)
@@ -2546,15 +2554,15 @@ static void getBaseType(sipSpec *pt, moduleDef *mod, classDef *defscope, argDef
         searchMappedTypes(pt,NULL,type);
 
         /*
-         * If we still have a template then see if we need to
-         * automatically instantiate it.
+         * If we still have a template then see if we need to automatically
+         * instantiate it.
          */
         if (type->atype == template_type)
         {
             mappedTypeTmplDef *mtt;
 
             for (mtt = pt->mappedtypetemplates; mtt != NULL; mtt = mtt->next)
-                if (sameScopedName(type->u.td->fqname, mtt->mt->type.u.td->fqname) && sameTemplateSignature(&type->u.td->types, &mtt->mt->type.u.td->types, TRUE))
+                if (sameScopedName(type->u.td->fqname, mtt->mt->type.u.td->fqname) && sameTemplateSignature(&mtt->mt->type.u.td->types, &type->u.td->types, TRUE))
                 {
                     type->u.mtd = instantiateMappedTypeTemplate(pt, mod, mtt, type);
                     type->atype = mapped_type;
@@ -2566,6 +2574,38 @@ static void getBaseType(sipSpec *pt, moduleDef *mod, classDef *defscope, argDef
 }
 
 
+/*
+ * If the type corresponds to a previously instantiated class template then
+ * replace it with the class that was created.
+ */
+static void resolveInstantiatedClassTemplate(sipSpec *pt, argDef *type)
+{
+    int a;
+    classDef *cd;
+    templateDef *td;
+    signatureDef *sd;
+
+    if (type->atype != template_type)
+        return;
+
+    td = type->u.td;
+    sd = &td->types;
+
+    for (a = 0; a < sd->nrArgs; ++a)
+        resolveInstantiatedClassTemplate(pt, &sd->args[a]);
+
+    for (cd = pt->classes; cd != NULL; cd = cd->next)
+        if (cd->td != NULL && sameScopedName(cd->td->fqname, td->fqname) &&
+            sameSignature(&cd->td->types, sd, TRUE))
+        {
+            type->atype = class_type;
+            type->u.cd = cd;
+
+            break;
+        }
+}
+
+
 /*
  * Instantiate a mapped type template and return it.
  */
@@ -2691,15 +2731,18 @@ void searchTypedefs(sipSpec *pt, scopedNameDef *snd, argDef *ad)
 {
     typedefDef *td;
 
-    for (td = pt -> typedefs; td != NULL; td = td -> next)
-        if (sameScopedName(td -> fqname,snd))
+    for (td = pt->typedefs; td != NULL; td = td->next)
+        if (sameScopedName(td->fqname, snd))
         {
             /* Copy the type. */
 
-            ad -> atype = td -> type.atype;
-            ad -> argflags |= td -> type.argflags;
-            ad -> nrderefs += td -> type.nrderefs;
-            ad -> u = td -> type.u;
+            ad->atype = td->type.atype;
+            ad->argflags |= td->type.argflags;
+            ad->nrderefs += td->type.nrderefs;
+            ad->u = td->type.u;
+
+            if (ad->original_type == NULL)
+                ad->original_type = td;
 
             break;
         }
@@ -2799,7 +2842,7 @@ static void ifaceFilesAreUsedByOverload(ifaceFileList **used, overDef *od)
         int a;
 
         for (a = 0; a < ta->nrArgs; ++a)
-            addToUsedList(used, ta->args[a]->cd->iff);
+            addToUsedList(used, ta->args[a]->iff);
     }
 }
 
@@ -3012,8 +3055,15 @@ static void assignEnumNrs(sipSpec *pt)
     enumDef *ed;
 
     for (ed = pt->enums; ed != NULL; ed = ed->next)
-        if (ed->fqcname != NULL)
-            ed->enumnr = ed->module->nrenums++;
+    {
+        if (ed->fqcname == NULL)
+            continue;
+
+        if (ed->ecd != NULL && isTemplateClass(ed->ecd))
+            continue;
+
+        ed->enumnr = ed->module->nrenums++;
+    }
 }
 
 
